#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>

#include "ember.h"
#include "avl.h"

/*#define DEBUG */

#ifdef DEBUG
#define _debug(args...) printf(args)
#else
#define _debug(args...)
#endif


/*************************************************************************
** Debugging routines
*/


/*
** _debugNodeValue : return a text representation of the key value
*/
static char *_debugNodeValue(tree,node)
	avl_tree	*tree;
	avl_node	*node;
{
	static char	buf[1024];

	if (!node)
	{
		strcpy(buf, "** NULL node **");
		return(buf);
	}
	switch(tree->key_type)
	{
		case AVL_KEY_INT:
			snprintf(buf, 1024, "%d", *(int*)node->key);
			return(buf);
			break;
		case AVL_KEY_UINT:
			snprintf(buf, 1024, "%u", *(uint*)node->key);
			return(buf);
			break;
		case AVL_KEY_STRING:
			return((char*)node->key);
			break;
	}
	strcpy(buf,"**UNKNOWN TYPE **");
	return(buf);
}



/*
** _printSubTree : 
*/
static void _printSubTree(tree, node, depth)
	avl_tree	*tree;
	avl_node	*node;
	int		depth;
{
	int		count;
	e_assoc		*tmp;

	if (!node)
	{
		return;
	}
	_printSubTree(tree, node->right, depth+1);
	count = depth;
	while(count)
	{
		printf("   ");
		count--;
	}
	tmp = (e_assoc*)node->data;
	printf("%s (%d) => '%s'\n", _debugNodeValue(tree, node), node->height,
		tmp->value.value.textVal);
	_printSubTree(tree, node->left, depth+1);
}
	





/*************************************************************************
** Private routines
*/


/*
** _maxHeight : 
*/

static int _maxHeight(node1, node2)
	avl_node	*node1,
			*node2;
{
	int		height1,
			height2;

	height1 = node1? node1->height : 0;
	height2 = node2? node2->height : 0;
	return ((height1 > height2)? height1 : height2);
}





/*
** _keyCompare : do a comparison of 2 keys based on the tree's key type
*/
static int _keyCompare(tree, key1, key2)
	avl_tree	*tree;
	void		*key1,
			*key2;
{
	switch(tree->key_type)
	{
		case AVL_KEY_INT:
			if (*(int*)key1 < *(int*)key2)
			{
				return(-1);
			}
			else if (*(int*)key1 > *(int*)key2)
			{
				return(1);
			}
			return(0);
			break;

		case AVL_KEY_UINT:
			if (*(uint*)key1 < *(uint*)key2)
			{
				return(-1);
			}
			else if (*(uint*)key1 > *(uint*)key2)
			{
				return(1);
			}
			return(0);
			break;

		case AVL_KEY_STRING:
			return(strcmp((char*)key1, (char*)key2));
			break;
	}
	return(0);
}




/*
** _moveNodeContents : move the key and data of a node to a different node
*/
static void _moveNodeContents(tree, src, dest)
	avl_tree	*tree;
	avl_node	*src,
			*dest;
{
	dest->data = src->data;
	src->data = NULL;
	switch(tree->key_type)
	{
		case AVL_KEY_INT:
		case AVL_KEY_UINT:
			bcopy(src->key_buf, dest->key_buf, sizeof(int));
			dest->key = dest->key_buf;
			break;

		case AVL_KEY_STRING:
			if (src->key == src->key_buf)
			{
				strcpy(dest->key_buf, src->key_buf);
				dest->key = dest->key_buf;
			}
			else
			{
				dest->key = strdup(src->key);
			}
			break;
	}
}
	


/*
** _newNode : create a new node containing the provided key and data
*/
static avl_node* _newNode(key, key_len, data)
	void		*key;
	int		key_len;
	void		*data;
{
	avl_node 	*node;

	/*
	** To reduce the numbers of malloc calls, there's a small buffer
	** in the node structure that we can use if the key is small
	** enought (ints and uints or small strings).  We trade off some
	** RAM for faster execution and that's fine for the intended use.
	*/

	node = (avl_node*) malloc(sizeof(avl_node));
	if (key_len < AVL_KEY_BUF_LEN)
	{
		node->key = node->key_buf;
	}
	else
	{
		node->key = (void*)malloc(key_len+1);
	}
	memcpy(node->key, key, key_len);
	node->left = NULL;
	node->right = NULL;
	node->height = 1; // new node is initially added at leaf
	node->data = data;
	return(node);
}

static void _freeNodeContents(tree, node)
	avl_tree	*tree;
	avl_node	*node;
{
	if (node->data)
		free(node->data);
	if (node->key != node->key_buf)
		free(node->key);
	node->data = node->key = NULL;
}

static void _freeNode(tree, node)
	avl_tree	*tree;
	avl_node	*node;
{
	if (node)
		free(node);
}




/*
** A utility function to right rotate subtree rooted with y
*/
static avl_node *_rightRotate(tree, y)
	avl_tree	*tree;
	avl_node	*y;
{
	avl_node	*x,
			*T2;

	_debug("\trotate right at '%s' height %d\n", _debugNodeValue(tree,y),
		y->height);
	x = y->left;
	T2 = x->right;

	// Perform rotation
	x->right = y;
	y->left = T2;

	// Update heights
	if (T2)
		T2->height = _maxHeight(T2->left, T2->right)+1;
	if (y)
		y->height = _maxHeight(y->left,y->right)+1;
	if (x)
		x->height = _maxHeight(x->left,x->right)+1;

	// Return new root
	return x;
}

/*
** A utility function to left rotate subtree rooted with x
** See the diagram given above.
*/
static avl_node *_leftRotate(tree, x)
	avl_tree	*tree;
	avl_node	*x;
{
	avl_node	*y,
			*T2;

	_debug("\trotate left at '%s' height %d\n", _debugNodeValue(tree,x),
		x->height);
	y = x->right;
	T2 = y->left;

	// Perform rotation
	y->left = x;
	x->right = T2;

	// Update heights
	if (T2)
		T2->height = _maxHeight(T2->left,T2->right)+1;
	if (x)
		x->height = _maxHeight(x->left,x->right)+1;
	if (y)
		y->height = _maxHeight(y->left,y->right)+1;

	// Return new root
	return y;
}


/*
** Get Balance factor of node N
*/

static int _getBalance(node)
	avl_node	*node;
{
	int		height_left,
			height_right;

	if (node == NULL)
		return 0;

	return( (node->left ? node->left->height : 0) -
		(node->right ? node->right->height : 0));
	/*
	return (_height(node->left) - _height(node->right));
	*/
	/*
	height_left = node->left ? node->left->height : 0;
	height_right = node->right ? node->right->height : 0;
	return(height_left - height_right);
	*/
	
}


/* 
** Given a non-empty binary search tree, return the node with minimum key 
** value found in that tree.
*/
avl_node *_minValueNode(tree, node)
	avl_tree	*tree;
	avl_node	*node;
{
	avl_node	*cur;

	/* loop down to find the leftmost leaf */
	cur = node;
	while (cur->left != NULL)
		cur = cur->left;
	return(cur);
}


/*
** Recursive function to insert key in subtree rooted
** with node and returns new root of subtree.
*/
static avl_node *_insertNode(tree, node, key, key_len, data)
	avl_tree	*tree;
	avl_node	*node;
	void		*key;
	int		key_len;
	void		*data;
{
	int		compare;

	if (node == NULL)
	{
		_debug("\tCreate new node\n");
		return(_newNode(key, key_len, data));
	}

	compare = _keyCompare(tree, key, node->key);
	_debug("\tCur node key '%s' height %d => %d\n",
		_debugNodeValue(tree,node), node->height, compare);

	/*
	** If we find an existing node with this key then just replace the
	** data. We aren't changing the tree structure at all so we can 
	** skip the height checks and rotations etc.
	*/
	if (compare == 0)
	{
		if(node->data)
			free(node->data);
		node->data = data;
		return(node);
	}

	/*
	** 
	*/
	if (compare < 0)
	{
		node->left=_insertNode(tree,node->left,key,key_len,data);
	}
	else 
	{
		node->right=_insertNode(tree,node->right,key,key_len,data);
	}

	/* 2. Update height of this ancestor node */
	node->height = _maxHeight(node->left,node->right) + 1;

	/* 3. Get the balance factor of this ancestor
		node to check whether this node became
		unbalanced */
	int balance = node ? _getBalance(node) : 0;

	// If this node becomes unbalanced, then
	// there are 4 cases

	// Left Left Case
	if (balance > 1 && _keyCompare(tree, key, node->left->key) < 0)
		return _rightRotate(tree, node);

	// Right Right Case
	if (balance < -1 && _keyCompare(tree, key, node->right->key) > 0)
		return _leftRotate(tree, node);

	// Left Right Case
	if (balance > 1 && _keyCompare(tree, key, node->left->key) > 0)
	{
		node->left = _leftRotate(tree, node->left);
		return _rightRotate(tree, node);
	}

	// Right Left Case
	if (balance < -1 && _keyCompare(tree, key, node->right->key) < 0)
	{
		node->right = _rightRotate(tree, node->right);
		return _leftRotate(tree, node);
	}

	/* return the (unchanged) node pointer */
	return node;
}



/*
** Recursive function to insert key in subtree rooted
** with node and returns new root of subtree.
*/
static avl_node *_findNode(tree, node, key)
	avl_tree	*tree;
	avl_node	*node;
	void		*key;
{
	int		compare;

	if (node == NULL)
	{
		/* Not found */
		return(NULL);
	}
	compare = _keyCompare(tree, key, node->key);
	_debug("\tCur node key '%s' height %d => %d\n",
		_debugNodeValue(tree,node), node->height, compare);

	if (compare == 0)
	{
		return(node);
	}
	if (compare < 0)
	{
		return(_findNode(tree, node->left, key));
	}
	else
	{
		return(_findNode(tree, node->right, key));
	}
}
		

static void _walkLeft(tree, state, cur_node)
	avl_tree	*tree;
	avl_walk	*state;
	avl_node	*cur_node;
{
	while(cur_node)
	{
		if (cur_node->left == NULL)
		{
			break;
		}
		state->node_stack[state->stack_depth] = cur_node;
		state->stack_depth++;
		cur_node = cur_node->left;
	}
	state->node = cur_node;
}



static void _walkRight(tree, state, cur_node)
	avl_tree	*tree;
	avl_walk	*state;
	avl_node	*cur_node;
{
	while(cur_node)
	{
		if (cur_node->right == NULL)
		{
			break;
		}
		state->node_stack[state->stack_depth] = cur_node;
		state->stack_depth++;
		cur_node = cur_node->right;
	}
	state->node = cur_node;
}



static void _destroyNode(tree, node)
	avl_tree	*tree;
	avl_node	*node;
{
	if (!node)
		return;

	_destroyNode(tree, node->left);
	_destroyNode(tree, node->right);
	_freeNodeContents(tree, node);
	_freeNode(tree, node);
}



static avl_node* _deleteNode(tree, node, key)
	avl_tree	*tree;
	avl_node	*node;
	void		*key;
{
	avl_node	*tmp;
	int		result,
			balance,
			balance_left,
			balance_right;

	if (node == NULL)
		return node;

	result = _keyCompare(tree, key,node->key);
	if (result < 0 )
	{
		node->left = _deleteNode(tree, node->left, key);
	}
	if (result > 0)
	{
		node->right = _deleteNode(tree, node->right, key);
	}
	if (result == 0)
	{
		/*
		** This is the node to delete
		*/
		tree->node_count--;
		if (node->left == NULL && node->right == NULL)
		{
			/*
			** The node has no children.  Just delete it
			*/
			_freeNodeContents(tree, node);
			_freeNode(tree, node);
			return(NULL);
		}

		if (node->left == NULL || node->right == NULL)
		{
			/*
			** This node has 1 child so just replace this
			** node with the kid
			*/
			if (node->left != NULL)
			{
				tmp = node->left;
			}
			else
			{
				tmp = node->right;
			}
			_freeNodeContents(tree, node);
			_freeNode(tree, node);
			node = tmp;
		}
		else
		{
			tmp = _minValueNode(tree, node->right);
			_freeNodeContents(tree, node);
			_moveNodeContents(tree, tmp, node);
			node->right = _deleteNode(tree, node->right, tmp->key);
		}
	}

	node->height = _maxHeight(node->left,node->right) + 1;
	balance = _getBalance(node);
	balance_left = node->left ? _getBalance(node->left) : 0;
	balance_right = node->right ? _getBalance(node->right) : 0;

	// Left Left Case
	if (balance > 1 && balance_left >= 0)
		return(_rightRotate(tree, node));

	// Left Right Case
	if (balance > 1 && balance_left < 0)
	{
		node->left = _leftRotate(tree, node->left);
		return(_rightRotate(tree, node));
	}

	// Right Right Case
	if (balance < -1 && balance_right <= 0)
		return(_leftRotate(tree,node));

	// Right Left Case
	if (balance < -1 && balance_right > 0)
	{
		node->right = _rightRotate(tree, node->right);
		return(_leftRotate(tree, node));
	}

	return node;
}


/***********************************************************************
** Public routines
***********************************************************************/



void avlPrintTree(tree)
	avl_tree	*tree;
{
	printf("\n");
	_printSubTree(tree, tree->root, 0);
	printf("\n");
}





avl_tree *avlCreate(type)
	int		type;
{
	avl_tree	*tree;

	tree = (avl_tree*)malloc(sizeof(avl_tree));
	tree->root = NULL;
	tree->key_type = type;
	tree->node_count = 0;
	return(tree);
}




int avlInsert(tree, key, data)
	avl_tree	*tree;
	void		*key,
			*data;
{
	int		key_len;

	switch(tree->key_type)
	{
		case AVL_KEY_INT:
			_debug("Insert key '%d'\n", *(int*)key);
			key_len = sizeof(int);
			break;
		case AVL_KEY_UINT:
			_debug("Insert key '%u'\n", *(uint*)key);
			key_len = sizeof(uint);
			break;
		case AVL_KEY_STRING:
			_debug("Insert key '%s'\n", (char*)key);
			key_len = strlen(key)+1;
			break;
		default:
			return(-1);
	}

	tree->root = _insertNode(tree, tree->root, key, key_len, data);
	tree->node_count++;

	return(0);
}



void avlDelete(tree, key)
	avl_tree	*tree;
	void		*key;
{
	avl_node	tmp;

	tmp.key = key;
	_debug("Delete key '%s'\n", _debugNodeValue(tree, &tmp));
	tree->root = _deleteNode(tree, tree->root, key);
}



void avlDestroy(tree)
	avl_tree	*tree;
{
	if (!tree)
		return;
	_destroyNode(tree, tree->root);
	free(tree);
}




int avlGetFirst(tree, state)
	avl_tree	*tree;
	avl_walk	*state;
{
	avl_node	*cur_node;

	/*
	** Can't do anything if we don't have a tree
	*/
	if (tree == NULL || tree->root == NULL)
	{
		return(0);
	}


	/*
	** Find the smallest value in the tree and keep track of
	** the nodes we pass through
	*/
	bzero(state, sizeof(avl_walk));
	cur_node = tree->root;
	_walkLeft(tree, state, cur_node);
	state->key = state->node->key;
	state->data = state->node->data;
	return(1);
}


int avlGetLast(tree, state)
	avl_tree	*tree;
	avl_walk	*state;
{
	avl_node	*cur_node;

	/*
	** Can't do anything if we don't have a tree
	*/
	if (tree == NULL || tree->root == NULL)
	{
		return(0);
	}


	/*
	** Find the largest value in the tree and keep track of
	** the nodes we pass through
	*/
	bzero(state, sizeof(avl_walk));
	cur_node = tree->root;
	_walkRight(tree, state, cur_node);
	state->key = state->node->key;
	state->data = state->node->data;
	return(1);
}



int avlGetNext(tree, state)
	avl_tree	*tree;
	avl_walk	*state;
{
	avl_node	*cur_node,
			*parent;


	/*
	** Find the node after the current node.  Is there a right subtree
	** off this node?  If so find the smallest value in that subtree
	*/

	cur_node = state->node;
	if (cur_node->right && state->prev_node != cur_node->right)
	{
		/* There's a right subtree off the current node */
		state->node_stack[state->stack_depth] = cur_node;
		state->stack_depth ++;
		cur_node = cur_node->right;
		_walkLeft(tree, state, cur_node);
		state->key = state->node->key;
		state->data = state->node->data;
		return(1);
	}


	/*
	** We've walked the right hand subtree of the current node.  Move
	** up the tree to the first node where we haven't walked the right
	** hand subtree
	*/
	while(cur_node)
	{
		if (state->stack_depth == 0)
		{
			return(0);
		}
		state->prev_node = cur_node;
		state->stack_depth--;
		cur_node = state->node_stack[state->stack_depth];
		if (state->prev_node == cur_node->left)
		{
			break;
		}
	}
	if (!cur_node)
	{
		return(0);
	}
	state->node = cur_node;
	state->key = state->node->key;
	state->data = state->node->data;
	return(1);
}

void *avlLookup(tree, key)
	avl_tree	*tree;
	void		*key;
{
	avl_node	*node;

	node = _findNode(tree, tree->root, key);
	if (node)
	{
		return(node->data);
	}
	return(NULL);
}
